/*
 *  Plasma.cpp - Plasma effect
 *               Written by Tinic Urou
 *               Public domain
 */

#include <AppKit.h>
#include <InterfaceKit.h>
#include <KernelKit.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>


// Types
typedef unsigned char 	UBYTE;
typedef unsigned short 	UWORD;
typedef unsigned long 	ULONG;
typedef char 			BYTE;
typedef short 			WORD;
typedef long 			LONG;

// Constants

const 	ULONG 			APP_SIGNATURE = '????';
const 	LONG 			DISPLAY_X = 280;
const 	LONG 			DISPLAY_Y = 200;

// Globals

		BBitmap			*my_bitmap;
		uchar			colors[256],*bits;
		uchar 			body[DISPLAY_X*DISPLAY_Y];	
		uchar 			tab1[(DISPLAY_X*2)*(DISPLAY_Y*2)];			
		uchar 			tab2[(DISPLAY_X*2)*(DISPLAY_Y*2)];			

// Window/Application messages

const 	LONG			MSG_REDRAW = 1;

class 					PlasmaWindow;

// Application object
class Plasma : public BApplication 
{
	public:
						Plasma();
			void 		ReadyToRun(void);
			bool 		QuitRequested(void);
			void 		AboutRequested(void);

	private:
	
			PlasmaWindow 	*the_window;
			BBitmap 	*the_bitmap;
};

struct thread_arg
{
	long x1;
	long y1;
	long x2;
	long y2;
	long x3;
	long y3;
	long x4;
	long y4;
	long roll;
	long start;
	long end;
};

// Window object
class PlasmaWindow : public BWindow 
{
	public:
	
						PlasmaWindow(short width, short height, BBitmap *bitmap);
			bool 		QuitRequested(void);
			void 		MessageReceived(BMessage *msg);

	private:
	
			void		recalc(void);
	static	void		draw_thread(struct thread_arg *a);

			BBitmap 	*the_bitmap;
			BView		*main_view;	
			ULONG		xmod;
};

// Create application object and start it
short main(short argc, char **argv)
{	
	Plasma *the_app;

	srand(system_time());

	the_app = new Plasma();
	the_app->Run();
	delete the_app;

	return 0;
}

/*
 *  Application constructor
 */

Plasma::Plasma() : BApplication(APP_SIGNATURE)
{
	the_window = NULL;
	the_bitmap = NULL;
}

/*
 *	Thats from me!
 */

void Plasma::AboutRequested(void) 
{
	char str[256];
	sprintf(str, "Plasma by Tinic Urou\n<5uro@informatik.uni-hamburg.de>\nFreely distributable.");
	BAlert *the_alert = new BAlert("", str, "OK");
	the_alert->Go();
}


/*
 *  Open window and start rendering
 */

void Plasma::ReadyToRun(void)
{
	// Allocate the bitmap		
	my_bitmap = the_bitmap = new BBitmap(BRect(0, 0, DISPLAY_X-1, DISPLAY_Y-1), B_COLOR_8_BIT);
	bits=(uchar *)the_bitmap->Bits();
	
	long i=0,j=0;
	float temp;
	while(i<DISPLAY_Y*2)
	{
		j=0;
		while(j<DISPLAY_X*2)
		{
			tab1[(i*DISPLAY_X*2)+j]=(uchar) (( sqrt(16.0+(DISPLAY_Y-i)*(DISPLAY_Y-i)+(DISPLAY_X-j)*(DISPLAY_X-j))-4) *5 );
			j++;
		}
		i++;
	}
	i=0;j=0;
	while(i<DISPLAY_Y*2)
	{
		j=0;
		while(j<DISPLAY_X*2)
		{
			temp=sqrt(16.0+(DISPLAY_Y-i)*(DISPLAY_Y-i)+(DISPLAY_X-j)*(DISPLAY_X-j))-4;
			tab2[(i*DISPLAY_X*2)+j]=(sin(temp/9.5)+1)*90;
			j++;
		}
		i++;
	}
		
	// Open window
	the_window = new PlasmaWindow(DISPLAY_X, DISPLAY_Y, the_bitmap);

	// Start action
	the_window->PostMessage(MSG_REDRAW);

	// Let it flow from the orginal to the distorted look
}


/*
 *  Quit requested (either by menu or by closing the window)
 */

bool Plasma::QuitRequested(void)
{
	// Make sure that the window closes first
	if (BApplication::QuitRequested()) 
	{	
		if (the_bitmap)	delete the_bitmap;
		return TRUE;
	}
	return FALSE;
}


/*
 *  Window constructor
 */

PlasmaWindow::PlasmaWindow(short width, short height, BBitmap *bitmap) : BWindow(BRect(0, 0, width-1, height-1), "Plasma by Tinic", B_TITLED_WINDOW, B_NOT_RESIZABLE | B_NOT_ZOOMABLE)
{
	the_bitmap = bitmap;
	
	xmod = bitmap->BytesPerRow();
	
	// Move window to right position
	MoveTo(80, 60);

	// Create bitmap view
	Lock();
	main_view = new BView(BRect(0, 0, DISPLAY_X-1, DISPLAY_Y-1),"", B_FOLLOW_NONE,NULL);
	AddChild(main_view);
	Unlock();

	// Show the window
	Show();
}


/*
 *  Closing the window quits the program
 */

bool PlasmaWindow::QuitRequested(void)
{
	be_app->PostMessage(B_QUIT_REQUESTED);
	return TRUE;
}


/*
 *  Handles redraw messages
 */

void PlasmaWindow::MessageReceived(BMessage *msg)
{
	switch (msg->what) 
	{
		case MSG_REDRAW:  // Redraw bitmap
			Lock();
			main_view->DrawBitmapAsync(the_bitmap);
			recalc();
			main_view->Sync();
			PostMessage(MSG_REDRAW);
			Unlock();
			break;
		default:
			BWindow::MessageReceived(msg);
	}
}

/*
 *  Main calculation
 */

void PlasmaWindow::draw_thread(struct thread_arg *a)
{
	long x1=a->x1;
	long y1=a->y1;
	long x2=a->x2;
	long y2=a->y2;
	long x3=a->x3;
	long y3=a->y3;
	long x4=a->x4;
	long y4=a->y4;
	long roll=a->roll;
	long start=a->start;
	long end=a->end;

	int k=DISPLAY_X*start;
	for(int i=start;i<end;i++)
	{
		for(int j=0;j<DISPLAY_X;j++)
		{
			bits[k+j]=
			colors
			[	
				(		
				tab1[(DISPLAY_X*2)*(i+y1)+j+x1]+roll+
				tab2[(DISPLAY_X*2)*(i+y2)+j+x2]+
				tab2[(DISPLAY_X*2)*(i+y3)+j+x3]+
				tab2[(DISPLAY_X*2)*(i+y4)+j+x4]
				)&0xFF
			];
		}
		k+=DISPLAY_X;
	}
}

#define mycol(u,a) (max(0.0,cos((u)+(a))))*255 
//#define mycol(u,a) ((cos((u)+(a))+1)*127)

void PlasmaWindow::recalc(void)
{
	static float	circle1=0,circle2=0,circle3=0,circle4=0,circle5=0,circle6=0,circle7=0,circle8=0;
	static short	roll=0;
	static short	i=0;
	long			returnvalue;
	thread_arg		arg1,arg2;
	thread_id		t1,t2;

	UBYTE 			*q=(UBYTE *)the_bitmap->Bits();
	static double 	r=1.0/6.0*M_PI,g=3.0/6.0*M_PI,b=5.0/6.0*M_PI;
	double 			u,v;
	
	for(int j=0;j<256;j++)
	{
		u=2*M_PI/256*j;
		colors[j]=index_for_color(mycol(u,r),mycol(u,g),mycol(u,b));
	}
	r+=0.05;
	g-=0.05;
	b+=0.1;
		
	i++;
	circle1+=0.085/6;
	circle2-=0.1/6;
	circle3+=.3/6;
	circle4-=.2/6;
	circle5+=.4/6;
	circle6-=.15/6;
	circle7+=.35/6;
	circle8-=.05/6;

	arg2.x2=arg1.x2=(DISPLAY_X/2)+(DISPLAY_X/2)*sin(circle1);
	arg2.y2=arg1.y2=(DISPLAY_Y/2)+(DISPLAY_Y/2)*cos(circle2);
	arg2.x1=arg1.x1=(DISPLAY_X/2)+(DISPLAY_X/2)*cos(circle3);
	arg2.y1=arg1.y1=(DISPLAY_Y/2)+(DISPLAY_Y/2)*sin(circle4);
	arg2.x3=arg1.x3=(DISPLAY_X/2)+(DISPLAY_X/2)*cos(circle5);
	arg2.y3=arg1.y3=(DISPLAY_Y/2)+(DISPLAY_Y/2)*sin(circle6);
	arg2.x4=arg1.x4=(DISPLAY_X/2)+(DISPLAY_X/2)*cos(circle7);
	arg2.y4=arg1.y4=(DISPLAY_Y/2)+(DISPLAY_Y/2)*sin(circle8);
	arg2.roll=arg1.roll=roll;
	arg1.start=0;
	arg2.start=DISPLAY_Y/2;
	arg1.end=DISPLAY_Y/2;
	arg2.end=DISPLAY_Y;
	
	resume_thread(t1=spawn_thread((thread_entry)draw_thread,"Plasma Draw Thread 1",B_NORMAL_PRIORITY,&arg1));
	resume_thread(t2=spawn_thread((thread_entry)draw_thread,"Plasma Draw Thread 2",B_NORMAL_PRIORITY,&arg2));
	wait_for_thread(t1,&returnvalue);
	wait_for_thread(t2,&returnvalue);
}
